#include <string>
#include <sstream>
#include <fstream>
#include <iostream>

#ifdef HAVE_CONFIG_H
#include "config.h" // to get VERSION
#endif

#include <arc/notify.h>
#include <arc/stringconv.h>
#include <arc/target.h>
#include <arc/jobsubmission.h>
#include <arc/jobftpcontrol.h>
#include <arc/mdsquery.h>
#include <arc/standardbrokers.h>
#include <arc/common.h>

JobSubmission::JobSubmission(Xrsl axrsl,
                             std::list<Target> targetlist,
                             bool adryrun) {
	xrsl = axrsl;
	targets = targetlist;
	dryrun = adryrun;
}


std::string JobSubmission::Submit() throw(XrslError, JobSubmissionError) {

	if (targets.size()==0)
		throw JobSubmissionError("No targets in target-list.");

	std::list<Target>::iterator targit;
	std::string jobid;
	for (targit = targets.begin(); targit != targets.end(); targit++) {

		notify(INFO) << "Queue " << targit->name << " at "
		             << targit->cluster.hostname << " selected." << std::endl;

		Xrsl jobxrsl;
		try {
			jobxrsl = PrepareXrsl(*targit);
		} catch(XrslError e) {
			throw JobSubmissionError(e.what());
		}

		if (dryrun && !jobxrsl.IsRelation("dryrun")) {
			XrslRelation dry("dryrun", operator_eq, "dryrun");
			jobxrsl.AddRelation(dry);
		}

		std::string contactstring = targit->cluster.contact;
		JobFTPControl ftpc;
		notify(DEBUG) << "Submitting xrsl: " << jobxrsl.str() << std::endl;

		try {
			jobid = ftpc.Submit(contactstring, jobxrsl.str(), localinputfiles);
		} catch(FTPControlError e) {
			notify(WARNING) << "Job submission failed to queue "
			                << targit->name << " at "
			                << targit->cluster.hostname << ": "
			                << e.what() << std::endl
			                << "Trying remaining queues." << std::endl;
			continue;
		}

		return jobid;
	}

	throw JobSubmissionError("Job submission failed.");
}


Xrsl JobSubmission::PrepareXrsl(Target& chosentarget) throw(XrslError) {

	Xrsl jobxrsl(chosentarget.GetXrsls().front());

	std::list<RuntimeEnvironment> middlewares =
	   chosentarget.cluster.middlewares;

	RuntimeEnvironment nordugridenv("nordugrid-unknown");
	std::list<RuntimeEnvironment>::iterator runit;
	for (runit = middlewares.begin(); runit!=middlewares.end(); runit++) {
		if (runit->Name()=="nordugrid") nordugridenv = *runit;
		break;
	}

	// rewrite xrsl so it's ready for jobsubmission
	XrslRelation rel1("queue", operator_eq, chosentarget.name);
	jobxrsl.AddRelation(rel1);
	XrslRelation rel2("action", operator_eq, "request");
	jobxrsl.AddRelation(rel2);
	XrslRelation rel3("savestate", operator_eq, "yes");
	jobxrsl.AddRelation(rel3);
	XrslRelation rel4("clientxrsl", operator_eq, xrsl.str());
	jobxrsl.AddRelation(rel4);

	char hostname[1024];
    globus_libc_gethostname(hostname, 1024);
	XrslRelation rel5("hostname", operator_eq, hostname);
	jobxrsl.AddRelation(rel5);

	std::string version("nordugrid-");
	version += VERSION;
	XrslRelation rel6("clientsoftware", operator_eq, version);
	jobxrsl.AddRelation(rel6);

	PrepareInputOutput(jobxrsl, "executable", "inputfiles");
	PrepareInputOutput(jobxrsl, "stdin", "inputfiles");
	PrepareInputOutput(jobxrsl, "stdout", "outputfiles");
	PrepareInputOutput(jobxrsl, "stderr", "outputfiles");
	PrepareInputOutput(jobxrsl, "gmlog", "outputfiles");

	// Add executable to executables
	XrslRelation rel7 = jobxrsl.GetRelation("executable");
	std::string exeval = rel7.GetSingleValue();

	std::list<std::string> exevals;
	if (jobxrsl.IsRelation("executables")) {
		exevals = jobxrsl.GetRelation("executables").GetListValue();
		jobxrsl.RemoveRelation("executables");
	}
	std::list<std::string>::iterator it;
	bool found = false;
	for (it = exevals.begin(); it != exevals.end(); it++) {
		if (*it==exeval) {
			found = true;
			break;
		}
	}
	if (found==false) exevals.push_front(exeval);

	XrslRelation rel8("executables", operator_eq, exevals);
	jobxrsl.AddRelation(rel8);

	bool addexetoargs = false;
	if (nordugridenv<RuntimeEnvironment("nordugrid-0.5.26"))
		addexetoargs = true;

	if (addexetoargs) {
		// Add executable to arguments

		std::list<std::string> argvals;
		if (jobxrsl.IsRelation("arguments")) {
			argvals = jobxrsl.GetRelation("arguments").GetListValue();
			jobxrsl.RemoveRelation("arguments");
		}
		found = false;
		for (it = argvals.begin(); it != argvals.end(); it++) {
			if (*it==exeval) {
				found = true;
				break;
			}
		}
		if (found==false) argvals.push_front(exeval);

		XrslRelation rel9("arguments", operator_eq, argvals);
		jobxrsl.AddRelation(rel9);

		jobxrsl.RemoveRelation("executable");
		XrslRelation rel10("executable", operator_eq, "/bin/echo");
		jobxrsl.AddRelation(rel10);
	}

	if (jobxrsl.IsRelation("join")) {
		XrslRelation joinrel = jobxrsl.GetRelation("join");
		std::string joinval = joinrel.GetSingleValue();
		if (joinval=="true" || joinval=="yes") {

			if (!jobxrsl.IsRelation("stdout"))
				throw XrslError("Xrsl-attribute \"join\" is set but "
				                "attribute \"stdout\" is not set.");
			if (jobxrsl.IsRelation("stderr"))
				throw XrslError("Xrsl-attribute \"join\" is set but "
				                "attribute \"stderr\" is also set.");

			XrslRelation out = jobxrsl.GetRelation("stdout");
			std::string outvalue = out.GetSingleValue();
			XrslRelation rel11("stderr", operator_eq, outvalue);
			jobxrsl.AddRelation(rel11);
		}
	}

	/*if (jobxrsl.IsRelation("stdin")) {
		XrslRelation stdin = jobxrsl.GetRelation("stdin");
		xrsl_operator op = stdin.GetOperator();
		std::string value = stdin.GetSingleValue();

		XrslRelation sstdin("sstdin", op, value);
		jobxrsl.RemoveRelation("stdin");
		jobxrsl.AddRelation(sstdin);
	}*/

	if (jobxrsl.IsRelation("starttime")) {
		XrslRelation starttime = jobxrsl.GetRelation("starttime");
		xrsl_operator op = starttime.GetOperator();

		try {
			Time start(starttime.GetSingleValue());
			XrslRelation newstarttime("starttime",
			                          op,
			                          TimeStamp(start, MDSTime));
			jobxrsl.RemoveRelation("starttime");
			jobxrsl.AddRelation(newstarttime);
		} catch(TimeError e) {
			throw XrslError("Illegal timeformat given in \"starttime\"");
		}
	}

	if (jobxrsl.IsRelation("lifetime")) {
		XrslRelation lifetime = jobxrsl.GetRelation("lifetime");
		xrsl_operator op = lifetime.GetOperator();
		unsigned long value = stringtoul(lifetime.GetSingleValue());

		std::stringstream ss;
		ss << value*60;
		XrslRelation newlifetime("lifetime", op, ss.str());
		jobxrsl.RemoveRelation("lifetime");
		jobxrsl.AddRelation(newlifetime);
	}

	// grap cputime
	long cputime = 0;
	long walltime = 0;
	if (jobxrsl.IsRelation("cputime") && jobxrsl.IsRelation("walltime")) {
		cputime = Seconds(jobxrsl.GetRelation("cputime").GetSingleValue());
		walltime = Seconds(jobxrsl.GetRelation("walltime").GetSingleValue());
	} else {
		cputime = chosentarget.GetCputime(jobxrsl);
		walltime = cputime;
	}

	if (cputime!=UNDEFINED) {
		if (nordugridenv<RuntimeEnvironment("nordugrid-0.5.27")) {
			cputime = (long)((cputime+59.0)/60.0);
			walltime = (long)((walltime+59.0)/60.0);
			// use minutes for pre 0.5.27 servers
		}

		if (jobxrsl.IsRelation("cputime"))
			jobxrsl.RemoveRelation("cputime");
		if (jobxrsl.IsRelation("walltime"))
			jobxrsl.RemoveRelation("walltime");
		if (jobxrsl.IsRelation("gridtime"))
			jobxrsl.RemoveRelation("gridtime");
		if (jobxrsl.IsRelation("benchmarks"))
			jobxrsl.RemoveRelation("benchmarks");

		std::stringstream ss1;
		ss1 << cputime;
		XrslRelation cpurel("cputime", operator_eq, ss1.str());
		jobxrsl.AddRelation(cpurel);

		std::stringstream ss2;
		ss2 << walltime;
		XrslRelation wallrel("walltime", operator_eq, ss2.str());
		jobxrsl.AddRelation(wallrel);
	}

	PrepareRE(jobxrsl, "runtimeenvironment", chosentarget);
	PrepareRE(jobxrsl, "middleware", chosentarget);
	PrepareRE(jobxrsl, "opsys", chosentarget);
	PrepareUpload(jobxrsl);

	return jobxrsl;
}


void JobSubmission::PrepareInputOutput(Xrsl& jobxrsl,
                                       const std::string& attr,
                                       const std::string& inoutattr) {

	if (!jobxrsl.IsRelation(attr)) return;
	std::string val = jobxrsl.GetRelation(attr).GetSingleValue();
	if (val[0]=='/' || val[0]=='$') return;

	std::list<std::list<std::string> > llstr;
	if (jobxrsl.IsRelation(inoutattr)) {
		llstr = jobxrsl.GetRelation(inoutattr).GetDoubleListValue();
		jobxrsl.RemoveRelation(inoutattr);
	}

	bool found = false;
	std::list<std::list<std::string> >::iterator it;
	for (it = llstr.begin(); it != llstr.end(); it++) {
		if (it->front()==val) {
			found = true;
			break;
		}
	}

	if (found==false) {
		std::list<std::string> newlist;
		newlist.push_front("");
		newlist.push_front(val);
		llstr.push_back(newlist);
	}
	
	XrslRelation rel(inoutattr, operator_eq, llstr);
	jobxrsl.AddRelation(rel);
}


void JobSubmission::PrepareUpload(Xrsl& jobxrsl) throw(XrslError) {

	std::list<std::list<std::string> > inputfiles;
	if (jobxrsl.IsRelation("inputfiles")) {
		inputfiles = jobxrsl.GetRelation("inputfiles").GetDoubleListValue();
	}

	std::list<std::list<std::string> > newinputfiles;

	std::list<std::list<std::string> >::iterator it;
	for (it = inputfiles.begin(); it != inputfiles.end(); it++) {
		std::string rfile = *(it->begin());
		std::string lfile = *(++it->begin());
		if (lfile.empty()) lfile = rfile;

		if (lfile.substr(0, 7)=="file://" ||
		    lfile.find("://")==std::string::npos) {

			if (lfile.substr(0, 7)=="file://") {
				lfile = lfile.substr(0, 7);
			} else {
				char path[PATH_MAX];
                getcwd(path, PATH_MAX);
                lfile = (std::string)path + '/' + lfile;
			}

			std::ifstream file(lfile.c_str());
			if (!file) throw XrslError("Could not open input file" +
			                           lfile);

			file.seekg(0, std::ios::end);
			size_t length = file.tellg();
			file.close();
			// missing: checksum evaluation

			std::stringstream ss;
			ss << length;

			std::list<std::string> newlist;
			newlist.push_back(rfile);
			newlist.push_back(ss.str());

			newinputfiles.push_back(newlist);
			localinputfiles[lfile] = rfile;
		} else {
			newinputfiles.push_back(*it);
		}
	}

	if (newinputfiles.size()>0) {
		XrslRelation newrel("inputfiles", operator_eq, newinputfiles);
		if (jobxrsl.IsRelation("inputfiles"))
			jobxrsl.RemoveRelation("inputfiles");
		jobxrsl.AddRelation(newrel);
	}

	return;
}


void JobSubmission::PrepareRE(Xrsl& jobxrsl,
                              const std::string& attr,
                              Target& chosentarget) throw(XrslError) {

	if (!jobxrsl.IsRelation(attr)) return;
	std::list<XrslRelation> allrel = jobxrsl.GetAllRelations(attr);

	std::list<RuntimeEnvironment> res;
	std::list<RuntimeEnvironment> clusterres;
	if (attr=="runtimeenvironment") {
		res = chosentarget.runtime_environments;
		clusterres = chosentarget.cluster.runtime_environments;
	} else if (attr=="middleware") {
		res = chosentarget.middlewares;
		clusterres = chosentarget.cluster.middlewares;
	} else if (attr=="opsys") {
		res = chosentarget.operating_systems;
		clusterres = chosentarget.cluster.operating_systems;
	} else {
		return;
	}

	if (res.size()>0) {
		res.insert(res.end(), clusterres.begin(), clusterres.end());
	} else {
		res = clusterres;
	}
	res.sort();
	res.unique();

	std::list<RuntimeEnvironment> chosenres;
	std::list<RuntimeEnvironment>::iterator it;
	for (it = res.begin(); it != res.end(); it++) {

		bool good = true;
		unsigned int sz = allrel.size();

		// this constructs the possible RE's under consideration.
		std::list<XrslRelation>::iterator relit;
		for (relit = allrel.begin(); relit != allrel.end(); relit++) {
			xrsl_operator op = relit->GetOperator();
			std::string value = relit->GetSingleValue();
			RuntimeEnvironment re(value);

			if (it->Name()!=re.Name()) {
				sz--;
				continue;
			}
			if (op==operator_eq && *it==re) continue;
			if (op==operator_lt && *it<re) continue;
			if (op==operator_gt && *it>re) continue;
			if (op==operator_lteq && *it<=re) continue;
			if (op==operator_gteq && *it>=re) continue;

			good = false; // if we reach this point, the RE is not valid.
		}

		if (good==true && sz!=0) chosenres.push_back(*it);
	}

	chosenres.sort();
	chosenres.unique();

	if (chosenres.size()>1) {
		std::list<RuntimeEnvironment>::iterator end = chosenres.end();
		end--;
		it = chosenres.begin();
		while (it != end && chosenres.size()>1) {
			if (it->Name()==(++it)->Name()) {
				--it;
				it = chosenres.erase(it);
			}
		}
	}

	while (jobxrsl.IsRelation(attr)) jobxrsl.RemoveRelation(attr);

	for (it = chosenres.begin(); it != chosenres.end(); it++) {
		XrslRelation rel(attr, operator_eq, it->str());
		jobxrsl.AddRelation(rel);
	}

	return;
}


std::list<Target> PrepareJobSubmission(Xrsl axrsl)
throw(MDSQueryError, XrslError, GIISError) {
	PerformXrslValidation(axrsl);
	std::list<Queue> queues = GetQueueInfo();
	std::list<Target> targets = ConstructTargets(queues, axrsl);
	PerformStandardBrokering(targets);
	return targets;
}


std::string SubmitJob(Xrsl axrsl, std::list<Target> targets, bool dryrun)
throw(JobSubmissionError, XrslError) {

	JobSubmission jobsub(axrsl, targets, dryrun);
	std::string jobid = jobsub.Submit();
	// Write jobid to file
	return jobid;
}
