#include "../../std.h"
#if 1
#include <sys/fsuid.h>
#endif

#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>
#endif

#define GRIDFTP_PLUGIN
#include "../../config/conf.h"
#include "../../jobs/job.h"
#include "../../jobs/commfifo.h"
#include "../../jobs/plugins.h"
#include "../userspec.h"
#include "../names.h"
#include "../misc.h"
#include "../../config/environment.h"
#include "../../config/config_pre.h"
#include "../../files/info_types.h"
#include "../../rsl/parse_rsl.h"
#include "../../misc/inttostring.h"
#include "../../misc/delete.h"
#include "../../misc/escaped.h"
#include "../../misc/log_time.h"
#include "../../transfer/proxy.h"
#include "../../run/run_parallel.h"
#include "../fileplugin/fileplugin.h"
#include "CertInfo.h"
#include "jobplugin.h"

RunParallel JobPlugin::run;


typedef struct {
  const JobUser* user;
  const std::string* job;
  const char* reason;
} job_subst_t;

#define IS_ALLOWED_READ  1
#define IS_ALLOWED_WRITE 2
#define IS_ALLOWED_LIST  4
#define IS_ALLOWED_RW    3
#define IS_ALLOWED_ALL   7

static void job_subst(std::string& str,void* arg) {
  job_subst_t* subs = (job_subst_t*)arg;
  if(subs->job) for(int p = 0;;) {
    p=str.find('%',p);
    if(p==std::string::npos) break;
    if(str[p+1]=='I') {
      str.replace(p,2,subs->job->c_str());
      p+=subs->job->length();
    } else if(str[p+1]=='S') {
      str.replace(p,2,"UNKNOWN"); // WRONG
      p+=7;
    } else if(str[p+1]=='O') {
      str.replace(p,2,subs->reason);
      p+=strlen(subs->reason);
    } else {
      p+=2;
    };
  };
  if(subs->user) subs->user->substitute(str);
}

// run external plugin to acquire non-unix local credentials
// U - user, J - job, O - reason
#define ApplyLocalCred(U,J,O) {                              \
  if(cred_plugin && (*cred_plugin)) {                        \
    job_subst_t subst_arg;                                   \
    subst_arg.user=U;                                        \
    subst_arg.job=J;                                         \
    subst_arg.reason=O;                                      \
    if(!cred_plugin->run(job_subst,&subst_arg)) {            \
      olog << "Failed to run plugin" << std::endl;                \
      return 1;                                              \
    };                                                       \
    if(cred_plugin->result() != 0) {                         \
      olog << "Plugin failed: " << cred_plugin->result() << std::endl;  \
      return 1;                                              \
    };                                                       \
  };                                                         \
}


JobPlugin::JobPlugin(std::istream &cfile,userspec_t &user_s) {
//  olog << "JobPlugin: constructor" << std::endl;
  initialized=true;
  rsl_opened=false;
  user=NULL;
  user_a=&(user_s.user);
  direct_fs=NULL;
  proxy_fname="";
  std::string configfile("");
  cont_plugins=NULL;
  cred_plugin=NULL;
  readonly=false;
  for(;;) {
    std::string rest;
    std::string command=config_read_line(cfile,rest);
    if(command.length() == 0) { break; } /* end of file - should not be here */
    else if(command == "configfile") {
      input_escaped_string(rest.c_str(),configfile); 
    } else if(command == "allownew") {
      std::string value("");
      input_escaped_string(rest.c_str(),value);
      if(strcasecmp(value.c_str(),"no") == 0) { readonly=true; }
      else if(strcasecmp(value.c_str(),"yes") == 0) { readonly=false; };
    } else if(command == "end") {
      break; /* end of section */
    }
    else {
      olog<<"Warning: unsupported configuration command: "<<command<<std::endl;
    };
  };
  if(!run.is_initialized()) {
    olog<<"Warning: Initialization of signal environment failed"<<std::endl;
  };
  if(!read_env_vars()) { initialized=false; }
  {
    if(configfile.length()) nordugrid_config_loc=configfile;
    user=new JobUser(user_s.get_uname());
    if(!user->is_valid()) { initialized=false; }
    else {
      /* read configuration */
      if(nordugrid_loc.length() != 0) {
        std::string session_root;
        std::string control_dir;
        std::string default_lrms;
        std::string default_queue;
        cont_plugins = new ContinuationPlugins;
        cred_plugin = new RunPlugin;
        std::string allowsubmit;
        bool strict_session;
        if(!configure_user_dirs(user_s.get_uname(),
          control_dir,session_root,
          default_lrms,default_queue,*cont_plugins,*cred_plugin,
          allowsubmit,strict_session)) {
          olog<<"Failed processing grid-manager configuration"<<std::endl;
          initialized=false;
        } else {
          user->SetControlDir(control_dir);
          user->SetSessionRoot(session_root);
          user->SetLRMS(default_lrms,default_queue);
          user->SetStrictSession(strict_session);
          for(;allowsubmit.length();) {
            std::string group = config_next_arg(allowsubmit);
            if(group.length()) readonly=true;
            if(user_a->check_group(group)) { readonly=false; break; };
          };
          if(readonly) olog<<"User is denied to submit new jobs."<<std::endl;
          /* link to the class for direct file access */
          std::string direct_config = "mount "+session_root+"\n";
          direct_config+="dir / nouser read cd dirlist delete append overwrite";
          direct_config+=" create "+
             inttostring(user->get_uid())+":"+inttostring(user->get_gid())+
             " 600:600";
          direct_config+=" mkdir "+
             inttostring(user->get_uid())+":"+inttostring(user->get_gid())+
             " 700:700\n";
          direct_config+="end\n";
#ifdef HAVE_SSTREAM
          std::stringstream fake_cfile(direct_config);
#else
          std::strstream fake_cfile;
          fake_cfile<<direct_config;
#endif
          direct_fs = new DirectFilePlugin(fake_cfile,user_s);
        };
      } else {
        initialized=false;
      };
    };
  };
  if(!initialized) if(user) { delete user; user=NULL; };
  if((user_s.user.proxy() == NULL) || (user_s.user.proxy()[0] == 0)) {
    olog<<"WARNING: no delegated credentials were passed"<<std::endl;
  } else {
    proxy_fname=user_s.user.proxy();
  };
  subject=user_s.user.DN();
  port=user_s.get_port();
  memcpy(host,user_s.get_host(),sizeof(host));
  job_id="";
#if 0
  if(user) {
    if(user->StrictSession()) {
      // Changing unix user. That means control directory must be
      // writable for every serviced user.
      user->SwitchUser(true);
    };
  };
#endif
}

JobPlugin::~JobPlugin(void) {
  delete_job_id();
  if(proxy_fname.length() != 0) { remove(proxy_fname.c_str()); };
  if(cont_plugins) delete cont_plugins;
  if(cred_plugin) delete cred_plugin;
//  olog << "JobPlugin: destructor " << count << std::endl;
};
 

int JobPlugin::makedir(std::string &dname) {
  if(!initialized) return 1;
  std::string id;
  bool spec_dir;
  if(is_allowed(dname.c_str(),true,&spec_dir,&id) & IS_ALLOWED_WRITE) {
    if(spec_dir) {
      error_description="Not allowed to write to special directory.";
      return 1;
    };
    ApplyLocalCred(user,&id,"write");
    if((getuid()==0) && (user) && (user->StrictSession())) {
      // Non-portable solution. Needed as long as Linux
      // does not support proper setuid in threads
      setfsuid(user->get_uid()); setfsgid(user->get_gid());
      int r=direct_fs->makedir(dname);
      setfsuid(getuid()); setfsgid(getgid());
      return r;
    };
    return direct_fs->makedir(dname);
  };
  error_description="Not allowed to this job.";
  return 1;
}

int JobPlugin::removefile(std::string &name) {
  if(!initialized) return 1;
  if(name.find('/') == std::string::npos) { /* request to cancel the job */
    if((name == "new") || (name == "info")) {
      error_description="Special directory can't be mangled.";
      return 1;
    };
    if(is_allowed(name.c_str()) & IS_ALLOWED_WRITE) {  /* owner of the job */
      JobId id(name); JobDescription job_desc(id,"");
      if(job_cancel_mark_put(job_desc,*user)) return 0;
    };
    error_description="Not allowed to cancel this job.";
    return 1;
  };
  const char* logname;
  std::string id;
  bool spec_dir;
  if(is_allowed(name.c_str(),false,&spec_dir,&id,&logname) & IS_ALLOWED_WRITE) {
    if(logname) {
      if((*logname) != 0) return 0; /* pretend status file is deleted */
    };
    if(spec_dir) {
      error_description="Special directory can't be mangled.";
      return 1; /* can delete status directory */
    };
    ApplyLocalCred(user,&id,"write");
    if((getuid()==0) && (user) && (user->StrictSession())) {
      // Non-portable solution. Needed as long as Linux
      // does not support proper setuid in threads
      setfsuid(user->get_uid()); setfsgid(user->get_gid());
      int r=direct_fs->removefile(name);
      setfsuid(getuid()); setfsgid(getgid());
      return r;
    };
    return direct_fs->removefile(name);
  };
  error_description="Not allowed to this job.";
  return 1;
}

int JobPlugin::removedir(std::string &dname) {
  if(!initialized) return 1;
  if(dname.find('/') == std::string::npos) { /* request to clean the job */
    if((dname == "new") || (dname == "info")) {
      error_description="Special directory can't be mangled.";
      return 1;
    };
    if(is_allowed(dname.c_str()) & IS_ALLOWED_WRITE) {  /* owner of the job */
      /* check the status */
      JobId id(dname); 
      job_state_t status=job_state_read_file(id,*user);
      if(status == JOB_STATE_FINISHED) { /* remove files */
        if(job_clean_final(JobDescription(id,user->SessionRoot()+"/"+id),
                           *user)) return 0;
      }
      else { /* put marks */
        JobDescription job_desc(id,"");
        bool res = job_cancel_mark_put(job_desc,*user);
        res &= job_clean_mark_put(job_desc,*user);
        if(res) return 0;
      };
      error_description="Failed to clean job.";
      return 1;
    };
    error_description="Not allowed to this job.";
    return 1;
  };
  std::string id;
  bool spec_dir;
  if(is_allowed(dname.c_str(),false,&spec_dir,&id) & IS_ALLOWED_WRITE) {
    if(spec_dir) {
      error_description="Special directory can't be mangled.";
      return 1;
    };
    ApplyLocalCred(user,&id,"write");
    if((getuid()==0) && (user) && (user->StrictSession())) {
      // Non-portable solution. Needed as long as Linux
      // does not support proper setuid in threads
      setfsuid(user->get_uid()); setfsgid(user->get_gid());
      int r=direct_fs->removedir(dname);
      setfsuid(getuid()); setfsgid(getgid());
      return r;
    };
    return direct_fs->removedir(dname);
  };
  error_description="Not allowed to this job.";
  return 1;
}

int JobPlugin::open(const char* name,open_modes mode,unsigned long long int size) {
  olog << "jobplugin: open: " << name << std::endl;
  if(!initialized) return 1;
  if(rsl_opened) {  /* unfinished request - cancel */
    olog<<"Request to open file with storing in progress"<<std::endl;
    rsl_opened=false;
    delete_job_id();
    error_description="Job submission is still in progress.";
    return 1; 
  };
  if( mode == GRIDFTP_OPEN_RETRIEVE ) {  /* open for reading */
    std::string fname;
    const char* logname;
    /* check if reading status files */
    bool spec_dir;
    if(is_allowed(name,false,&spec_dir,&fname,&logname) & IS_ALLOWED_READ) {
      if(logname) {
        if((*logname) != 0) {
          fname=user->ControlDir()+"/job."+fname+"."+logname;
          return direct_fs->open_direct(fname.c_str(),mode);
        };
      };
      if(spec_dir) {
        error_description="Special directory can't be mangled.";
        return 1;
      };
      ApplyLocalCred(user,&fname,"read");
      if((getuid()==0) && (user) && (user->StrictSession())) {
        // Non-portable solution. Needed as long as Linux 
        // does not support proper setuid in threads
        setfsuid(user->get_uid()); setfsgid(user->get_gid());
        int r=direct_fs->open(name,mode);
        setfsuid(getuid()); setfsgid(getgid());
        return r;
      };
      return direct_fs->open(name,mode);
    };
    error_description="Not allowed to this job.";
    return 1;
  }
  else if( mode == GRIDFTP_OPEN_STORE ) {
    std::string name_f(name);
    int n = name_f.find('/');
    if((n != std::string::npos) && (n != 0)) {
      if(((n==3) && 
          (strncmp(name_f.c_str(),"new",n) == 0)
         ) ||
         ((n==job_id.length()) && 
          (strncmp(name_f.c_str(),job_id.c_str(),n) == 0)
         )) {
        if(name_f.find('/',n+1) != std::string::npos) { // should not contain subdirs
          error_description="Can't create subdirectory here.";
          return 1;
        };
        if(job_id.length() == 0) {
          if(!make_job_id()) {
            olog<<"Failed to allocate id for job"<<std::endl;
            error_description="Failed to allocate ID for job.";
            return 1;
          };
        };
        olog<<"Accepting submission of new job: "<<job_id<<std::endl;
        memset(job_rsl,0,sizeof(job_rsl));
        rsl_opened=true;
        return 0;
      };
    };
    std::string id;
    bool spec_dir;
    if(is_allowed(name,true,&spec_dir,&id) & IS_ALLOWED_WRITE) {
      if(spec_dir) {
        error_description="Special directory can't be mangled.";
        return 1;
      };
      ApplyLocalCred(user,&id,"write");
      if((getuid()==0) && (user) && (user->StrictSession())) {
        // Non-portable solution. Needed as long as Linux
        // does not support proper setuid in threads
        setfsuid(user->get_uid()); setfsgid(user->get_gid());
        int r=direct_fs->open(name,mode,size);
        setfsuid(getuid()); setfsgid(getgid());
        return r;
      };
      return direct_fs->open(name,mode,size);
    };
    error_description="Not allowed to this job.";
    return 1;
  }
  else {
    olog << "ERROR: unknown open mode " << mode << std::endl;
    error_description="Unknown/unsupported request.";
    return 1;
  };
}

int JobPlugin::close(bool eof) {
  if(!initialized) return 1;
  if(!rsl_opened) {
    if((getuid()==0) && (user) && (user->StrictSession())) {
      // Non-portable solution. Needed as long as Linux
      // does not support proper setuid in threads
      setfsuid(user->get_uid()); setfsgid(user->get_gid());
      int r=direct_fs->close(eof);
      setfsuid(getuid()); setfsgid(getgid());
      return r;
    };
    return direct_fs->close(eof);
  };
  rsl_opened=false;
  if(job_id.length() == 0) {
    error_description="There is no job ID defined.";
    return 1;
  };
  if(!eof) { delete_job_id(); return 0; }; /* download was canceled */
  /* *************************************************
   * store RSL (description)                         *
   ************************************************* */
  std::string rsl_fname=user->ControlDir()+"/job."+job_id+".description";
  std::string acl("");
  {
    int l;
    /* int h=::open(rsl_fname.c_str(),O_WRONLY | O_CREAT | O_EXCL,0600); */
    int h=::open(rsl_fname.c_str(),O_WRONLY,0600);
    fix_file_owner(rsl_fname,*user);
    if(h == -1) { l=-1; }
    else {
      const char* s;
      for(l=0,s=job_rsl;(*s) && (l!=-1);s+=l) l=::write(h,s,strlen(s));
      ::close(h);
    };
    if(l == -1) { 
      olog<<"Failed writing RSL"<<std::endl;
      error_description="Failed to store job RSL description.";
      delete_job_id(); return 1;
    };
  };
  /* analyze rsl (checking, substituting, etc)*/
  JobLocalDescription job_desc;
  if(!parse_rsl(rsl_fname.c_str(),job_desc,&acl)) {
    olog<<"Failed parsing RSL"<<std::endl;
    error_description="Failed to parse job RSL description.";
    delete_job_id();
    return 1;
  };
  if(job_desc.action.length() == 0) job_desc.action="request";
  if(job_desc.action == "cancel") {
    delete_job_id();
    if(job_desc.jobid.length() == 0) {
      error_description="Missing ID in request to cancel job.";
      olog<<error_description<<std::endl;
      return 1;
    };
    return removefile(job_desc.jobid);
  };
  if(job_desc.action == "clean") {
    delete_job_id();
    if(job_desc.jobid.length() == 0) {
      error_description="Missing ID in request to clean job.";
      olog<<error_description<<std::endl;
      return 1;
    };
    return removedir(job_desc.jobid);
  };
  if(job_desc.action == "renew") {
    delete_job_id();
    if(job_desc.jobid.length() == 0) {
      error_description="Missing ID in request to renew credentials.";
      olog<<error_description<<std::endl;
      return 1;
    };
    return checkdir(job_desc.jobid);
  };
  if(job_desc.action == "restart") {
    delete_job_id();
    if(job_desc.jobid.length() == 0) {
      error_description="Missing ID in request to clean job.";
      olog<<error_description<<std::endl;
      return 1;
    };
    const char* logname;
    std::string id;
    if(!(is_allowed(job_desc.jobid.c_str(),false,NULL,&id,&logname) & 
                                                       IS_ALLOWED_LIST)) {
      error_description="Not allowed to this job.";
      olog<<error_description<<std::endl;
      return 1;
    };
    if(job_desc.jobid != id) {
      error_description="Wrong ID specified.";
      olog<<error_description<<std::endl;
      return 1;
    };
    JobLocalDescription job_desc;
    if(!job_local_read_file(id,*user,job_desc)) {
      error_description="Job is probably corrupted: can't read internal information.";
      olog<<error_description<<std::endl;
      return 1;
    };
    if(job_desc.failedstate.length() == 0) {
      error_description="Job can't be restarted.";
      olog<<error_description<<std::endl;
      return 1;
    };
    if(job_desc.reruns <= 0) {
      error_description="Job can't be restarted anymore.";
      olog<<error_description<<std::endl;
      return 1;
    };
    if(!job_restart_mark_put(JobDescription(id,""),*user)) {
      error_description="Failed to report restart request.";
      olog<<error_description<<std::endl;
      return 1;
    };
    return 0;
  };
  if(job_desc.action != "request") {
    olog<<"action("<<job_desc.action<<") != request"<<std::endl;
    error_description="Wrong action in job RSL description.";
    delete_job_id();
    return 1;
  };
  if(job_desc.queue.length() == 0) job_desc.queue=user->DefaultQueue();
  if(job_desc.lrms.length() == 0) job_desc.lrms=user->DefaultLRMS();
  if((job_desc.jobid.length() != 0) && (job_desc.jobid != job_id)) {
    delete_job_id(); 
    if(!make_job_id(job_desc.jobid)) {
      olog<<"Failed allocating requested job id: "<<job_desc.jobid<<std::endl;
      remove(rsl_fname.c_str());
      error_description="Failed to allocate requested job ID.";
      return 1;
    };
    rsl_fname=user->ControlDir()+"/job."+job_id+".description";
    {
      int l;
      int h=::open(rsl_fname.c_str(),O_WRONLY,0600);
      fix_file_owner(rsl_fname,*user);
      if(h == -1) { l=-1; }
      else {
        const char* s;
        for(l=0,s=job_rsl;(*s) && (l!=-1);s+=l) l=::write(h,s,strlen(s));
        ::close(h);
      };
      if(l == -1) {
        olog<<"Failed writing RSL"<<std::endl;
        remove(rsl_fname.c_str());
        error_description="Failed to store job RSL description.";
        delete_job_id(); return 1;
      };
    };
  };
  if(!preprocess_rsl(rsl_fname,user->SessionRoot()+"/"+job_id,job_id)) {
    olog<<"Failed preprocessing RSL"<<std::endl;
    error_description="Failed to preprocess job RSL description.";
    delete_job_id(); 
    return 1;
  };
  /* ****************************************
   * Start local file                       *
   **************************************** */
  /* !!!!! some parameters are unchecked here - rerun,diskspace !!!!! */
/*
  job_desc.lrms=lrms;
  job_desc.queue=queue;
*/
  job_desc.jobid=job_id;
  job_desc.starttime=time(NULL);
  job_desc.DN=subject;
  if(port != 0) {
    job_desc.clientname=
       inttostring(host[0])+"."+inttostring(host[1])+"."+
       inttostring(host[2])+"."+inttostring(host[3])+":"+
       inttostring(port);
  };
  /* ***********************************************
   * Try to create proxy                           *
   *********************************************** */
  if(proxy_fname.length() != 0) {
    std::string fname=user->ControlDir()+"/job."+job_id+".proxy";
    int h=::open(fname.c_str(),O_WRONLY | O_CREAT | O_EXCL,0600);
    if(h == -1) {
      error_description="Failed to store credentials.";
      return 1;
    };
    int hh=::open(proxy_fname.c_str(),O_RDONLY);
    if(hh == -1) {
      ::close(h);
      ::remove(fname.c_str());
      error_description="Failed to read credentials.";
      return 1;
    };
    fix_file_owner(fname,*user);
    int l,ll;
    const char* s;
    char buf[256]; 
    for(;;) {
      ll=::read(hh,buf,sizeof(buf));
      if((ll==0) || (ll==-1)) break;
      for(l=0,s=buf;(ll>0) && (l!=-1);s+=l,ll-=l) l=::write(h,s,ll);
      if(l==-1) break;
    };
    ::close(h);
    ::close(hh);
    CertInfo ci(fname.c_str());
    job_desc.expiretime = (ci.TimeLeft()+time(NULL));
  };
  /* ******************************************
   * Write local file                         *
   ****************************************** */
  JobDescription job(job_id,"",JOB_STATE_ACCEPTED);
  if(!job_local_write_file(job,*user,job_desc)) {
    olog<<"Failed writing local description"<<std::endl;
    delete_job_id(); 
    error_description="Failed to create job description.";
    return 1;
  };
  if(acl.length() != 0) {
    if(!job_acl_write_file(job_id,*user,acl)) {
      olog<<"Failed writing ACL"<<std::endl;
      delete_job_id(); 
      error_description="Failed to process/store job ACL.";
      return 1;
    };
  };
  /* ***********************************************
   * Call authentication/authorization plugin/exec *
   *********************************************** */
  int result;
  std::string response;
  /* talk to external plugin to ask if we can proceed */
  ContinuationPlugins::action_t act =
                 cont_plugins->run(job,*user,result,response);
  // analyze result
  if(act == ContinuationPlugins::act_fail) {
    olog<<"Failed to run external plugin: "<<response<<std::endl;
    delete_job_id();
    error_description="Job is not allowed by external plugin: "+response;
    return 1;
  } else if(act == ContinuationPlugins::act_log) {
    // Scream but go ahead
    olog<<"Failed to run external plugin: "<<response<<std::endl;
  } else if(act == ContinuationPlugins::act_pass) {
    // Just continue
    if(response.length()) olog<<"Plugin response: "<<response<<std::endl;
  } else {
    olog<<"Failed to run external plugin"<<std::endl;
    delete_job_id();
    error_description="Failed to pass external plugin.";
    return 1;
  };
  /* ************************************************************
   * From here code accesses filesystem on behalf of local user *
   ************************************************************ */
  if(cred_plugin && (*cred_plugin)) {
    job_subst_t subst_arg;
    subst_arg.user=user;
    subst_arg.job=&job_id;
    subst_arg.reason="new";
    // run external plugin to acquire non-unix local credentials
    if(!cred_plugin->run(job_subst,&subst_arg)) {
      olog << "Failed to run plugin" << std::endl;
      delete_job_id();
      error_description="Failed to obtain external credentials.";
      return 1;
    };
    if(cred_plugin->result() != 0) {
      olog << "Plugin failed: " << cred_plugin->result() << std::endl;
      delete_job_id();
      error_description="Failed to obtain external credentials.";
      return 1;
    };
  };
  /* *******************************************
   * Create session directory                  *
   ******************************************* */
  std::string dir=user->SessionRoot()+"/"+job_id;
  // Non-portable solution. Needed as long as Linux
  // does not support proper setuid in threads
  if((getuid()==0) && (user) && (user->StrictSession())) {
    setfsuid(user->get_uid()); setfsgid(user->get_gid());
  };
  if(mkdir(dir.c_str(),0700) != 0) {
    if((getuid()==0) && (user) && (user->StrictSession())) {
      setfsuid(getuid()); setfsgid(getgid());
    };
    olog<<"Failed to create session directory"<<std::endl;
    delete_job_id();
    error_description="Failed to create session directory.";
    return 1;
  };
  if((getuid()==0) && (user) && (user->StrictSession())) {
    setfsuid(getuid()); setfsgid(getgid());
  };
  fix_file_owner(dir,*user);
  /* **********************************************************
   * Create status file (do it last so GM picks job up here)  *
   ********************************************************** */
  if(!job_state_write_file(job,*user,JOB_STATE_ACCEPTED)) {
    olog<<"Failed writing status"<<std::endl;
    delete_job_id(); 
    error_description="Failed registering job in grid-manager.";
    return 1;
  };
  SignalFIFO(*user);
  job_id.resize(0);
  return 0;
}

int JobPlugin::read(unsigned char *buf,unsigned long long int offset,unsigned long long int *size) {
  if(!initialized) {
    error_description="Transfer is not initialised.";
    return 1;
  };
  error_description="Failed to read from disc.";
  if((getuid()==0) && (user) && (user->StrictSession())) {
    // Non-portable solution. Needed as long as Linux
    // does not support proper setuid in threads
    setfsuid(user->get_uid()); setfsgid(user->get_gid());
    int r=direct_fs->read(buf,offset,size);
    setfsuid(getuid()); setfsgid(getgid());
    return r;
  };
  return direct_fs->read(buf,offset,size);
}

int JobPlugin::write(unsigned char *buf,unsigned long long int offset,unsigned long long int size) {
  if(!initialized) {
    error_description="Transfer is not initialised.";
    return 1;
  };
  error_description="Failed to write to disc.";
  if(!rsl_opened) {
    if((getuid()==0) && (user) && (user->StrictSession())) {
      // Non-portable solution. Needed as long as Linux
      // does not support proper setuid in threads
      setfsuid(user->get_uid()); setfsgid(user->get_gid());
      int r=direct_fs->write(buf,offset,size);
      setfsuid(getuid()); setfsgid(getgid());
      return r;
    };
    return direct_fs->write(buf,offset,size);
  };
  /* write to rsl */
  if(job_id.length() == 0) {
    error_description="No job ID defined.";
    return 1;
  };
  if((offset >= (sizeof(job_rsl)-1)) ||
     (size >= (sizeof(job_rsl)-1)) ||
     ((offset+size) >= (sizeof(job_rsl)-1)) ) {
    error_description="RSL is too big.";
    return 1;
  };
  memcpy(job_rsl,buf+offset,size);
  return 0;
}

int JobPlugin::readdir(const char* name,std::list<DirEntry> &dir_list,DirEntry::object_info_level mode) {
  if(!initialized) {
    error_description="Plugin is not initialised.";
    return 1;
  };
  if((name[0] == 0) || (strcmp("info",name) == 0)) { /* root jobs directory or jobs' info */
    if(name[0] == 0) {
      DirEntry dent_new(false,"new");
      DirEntry dent_info(false,"info");
      dent_new.may_dirlist=true;
      dent_info.may_dirlist=true;
      dir_list.push_back(dent_new);
      dir_list.push_back(dent_info);
    };
    struct dirent file_;
    struct dirent *file;
    std::string cdir=user->ControlDir();
    DIR *dir=opendir(cdir.c_str());
    if(dir != NULL) for(;;) {
      readdir_r(dir,&file_,&file);
      if(file == NULL) break;
      int l=strlen(file->d_name);
      if(l>15) {
        if(!strncmp(file->d_name,"job.",4)) {
          if(!strncmp((file->d_name)+(l-6),".local",6)) { /* can only process those with .local information */
            JobLocalDescription job_desc;
            std::string fname=cdir+'/'+file->d_name;
            if(job_local_read_file(fname,job_desc)) {
              if(job_desc.DN == subject) {
                JobId id((file->d_name)+4,l-6-4);
                dir_list.push_back(DirEntry(false,id.c_str()));
              };
            };
          };
        };
      };
    };
    closedir(dir);
    return 0;
  };
  if(strcmp(name,"new") == 0) { /* directory /new is always empty */
    return 0;
  };
  /* check for allowed job directory */
  const char* logname;
  std::string id;
  std::string log;
  if(is_allowed(name,false,NULL,&id,&logname,&log) & IS_ALLOWED_LIST) {
    if(logname) {
      if((*logname) != 0) {
        if(strchr(logname,'/') != NULL) return 1; /* no subdirs */
        id=user->ControlDir()+"/job."+id+"."+logname;
        struct stat64 st;
        if(::stat64(id.c_str(),&st) != 0) return 1;
        if(!S_ISREG(st.st_mode)) return 1;
        DirEntry dent(true,logname);
        if(strncmp(logname,"proxy",5) != 0) dent.may_read=true;
        dir_list.push_back(dent);
        return -1;
      };
      DIR* d=::opendir(user->ControlDir().c_str());
      if(d == NULL) { return 1; }; /* maybe return ? */
      struct dirent *de;
      struct dirent de_;
      id="job."+id+".";
      int idl = id.length();
      for(;;) {
        readdir_r(d,&de_,&de);
        if(de == NULL) break;
        if((!strcmp(de->d_name,".")) || (!strcmp(de->d_name,".."))) continue;
        if(strncmp(id.c_str(),de->d_name,idl) != 0) continue;
        DirEntry dent(true,(de->d_name)+idl);
        if(strncmp((de->d_name)+idl,"proxy",5) != 0) dent.may_read=true;
        dir_list.push_back(dent);
      };
      ::closedir(d);
      return 0;
    };
    if(log.length() > 0) {
      char* s = strchr(name,'/');
      if((s == NULL) || (s[1] == 0)) {
        DirEntry dent(false,log.c_str());
        dent.may_dirlist=true;
        dir_list.push_back(dent);
      };
    };
    /* allowed - pass to file system */
    ApplyLocalCred(user,&id,"read");
    if((getuid()==0) && (user) && (user->StrictSession())) {
      // Non-portable solution. Needed as long as Linux
      // does not support proper setuid in threads
      setfsuid(user->get_uid()); setfsgid(user->get_gid());
      int r=direct_fs->readdir(name,dir_list,mode);
      setfsuid(getuid()); setfsgid(getgid());
      return r;
    };
    return direct_fs->readdir(name,dir_list,mode);
  };
  error_description="Not allowed to this job.";
  return 1;
}

int JobPlugin::checkdir(std::string &dirname) {
  if(!initialized) return 1;
  /* chdir to /new will create new job */
  if(dirname.length() == 0) return 0; /* root */
  if(dirname == "new") { /* new job */
    if(!make_job_id()) {
      error_description="Failed to allocate ID for job.";
      return 1;
    };
    dirname=job_id;
    return 0;
  };
  if(dirname == "info") { /* always allowed */
    return 0;
  };
  const char* logname;
  std::string id;
  if(is_allowed(dirname.c_str(),false,NULL,&id,&logname) & IS_ALLOWED_LIST) {
    if(logname) {
      if((*logname) != 0) {
        error_description="There is no such special subdirectory.";
        return 1; /* log directory has no subdirs */
      };
      return 0;
    };
    if((dirname == id) && (proxy_fname.length())) {  /* cd to session directory - renew proxy request */
      std::string old_proxy_fname=user->ControlDir()+"/job."+id+".proxy";
      CertInfo new_ci(proxy_fname.c_str());
      time_t new_proxy_expires=new_ci.TimeLeft()+time(NULL);
      CertInfo old_ci(old_proxy_fname.c_str());
      time_t old_proxy_expires;
      if(!old_ci) { old_proxy_expires=time(NULL); }
      else { old_proxy_expires=old_ci.TimeLeft()+time(NULL); }
      if(new_proxy_expires > old_proxy_expires) {
        if(renew_proxy(old_proxy_fname.c_str(),proxy_fname.c_str()) == 0) {
          olog<<"New proxy expires at "<<mds_time(new_proxy_expires)<<std::endl;
          JobLocalDescription job_desc;
          if(job_local_read_file(id,*user,job_desc)) {
            JobDescription job(id,"",JOB_STATE_ACCEPTED);
            job_desc.expiretime=new_proxy_expires;
            if(!job_local_write_file(job,*user,job_desc)) {
              olog<<"Failed to write 'local' information"<<std::endl;
            };
          };
          if(!job_restart_mark_put(JobDescription(id,""),*user)) {
            olog<<"Failed to report renewed proxy to job"<<std::endl;
          };
          ApplyLocalCred(user,&id,"renew");
          if(cred_plugin && (*cred_plugin)) {
            job_subst_t subst_arg;
            subst_arg.user=user;
            subst_arg.job=&id;
            subst_arg.reason="renew";
            if(!cred_plugin->run(job_subst,&subst_arg)) {
              olog << "Failed to run plugin" << std::endl; return 1;
            };
            if(cred_plugin->result() != 0) {
              olog << "Plugin failed: " << cred_plugin->result() << std::endl;
              error_description="External credentials failed.";
              return 1;
            };
          };
        } else {
          olog<<"Failed to renew proxy"<<std::endl;
        };
      };
    };
    ApplyLocalCred(user,&id,"read");
    if((getuid()==0) && (user) && (user->StrictSession())) {
      // Non-portable solution. Needed as long as Linux
      // does not support proper setuid in threads
      setfsuid(user->get_uid()); setfsgid(user->get_gid());
      int r=direct_fs->checkdir(dirname);
      setfsuid(getuid()); setfsgid(getgid());
      return r;
    };
    return direct_fs->checkdir(dirname);
  };
  error_description="Not allowed to this job.";
  return 1;
}

int JobPlugin::checkfile(std::string &name,DirEntry &info,DirEntry::object_info_level mode) {
  if(!initialized) return 1;
  if(name.length() == 0) {
    info.name=""; info.is_file=false;
    return 0; 
  };
  if((name == "new") || (name == "info")) {
    info.name=""; info.is_file=false;
    return 0; 
  };
  const char* logname;
  std::string id;
  if(is_allowed(name.c_str(),false,NULL,&id,&logname) & IS_ALLOWED_LIST) {
    if(logname) {
      if((*logname) == 0) { /* directory itself */
        info.is_file=false; info.name=""; info.may_dirlist=true;
      }
      else {
        id=user->ControlDir()+"/job."+id+"."+logname;
        struct stat64 st;
        if(::stat64(id.c_str(),&st) != 0) {
          error_description="There is no such special file.";
          return 1;
        };
        if(!S_ISREG(st.st_mode)) {
          error_description="There is no such special file.";
          return 1;
        };
        info.is_file=true; info.name=""; info.may_read=true;
      };
      return 0;
    };
    ApplyLocalCred(user,&id,"read");
    if((getuid()==0) && (user) && (user->StrictSession())) {
      // Non-portable solution. Needed as long as Linux
      // does not support proper setuid in threads
      setfsuid(user->get_uid()); setfsgid(user->get_gid());
      int r=direct_fs->checkfile(name,info,mode);
      setfsuid(getuid()); setfsgid(getgid());
      return r;
    };
    return direct_fs->checkfile(name,info,mode);
  };
  error_description="Not allowed to this job.";
  return 1;
}

bool JobPlugin::delete_job_id(void) {
  if(job_id.length() != 0) {
    job_clean_final(JobDescription(job_id,user->SessionRoot()+"/"+job_id),*user);
    job_id="";
  };
  return true;
}

bool JobPlugin::make_job_id(const std::string &id) {
  if(readonly) {
    olog<<"Not accepting new jobs"<<std::endl;
    return false;
  };
  if((id.find('/') != std::string::npos) || (id.find('\n') != std::string::npos)) {
    olog<<"ID contains forbidden characters"<<std::endl;
    return false;
  };
  if((id == "new") || (id == "info")) return false;
  // claim id by creating empty description file
  std::string fname=user->ControlDir()+"/job."+id+".description";
  struct stat st;
  if(stat(fname.c_str(),&st) == 0) return false;
  int h = ::open(fname.c_str(),O_RDWR | O_CREAT | O_EXCL,S_IRWXU);
  // So far assume control directory is on local fs.
  // TODO: add locks or links for NFS
  if(h == -1) return false;
  fix_file_owner(fname,*user);
  close(h);
  delete_job_id();
  job_id=id;
  return true;
}

bool JobPlugin::make_job_id(void) {
  if(readonly) {
    olog<<"Not accepting new jobs"<<std::endl;
    return false;
  };
  int i;
  delete_job_id();
  for(i=0;i<100;i++) {
    job_id=inttostring((unsigned int)getpid())+
           inttostring((unsigned int)time(NULL))+
           inttostring(rand(),1);
    std::string fname=user->ControlDir()+"/job."+job_id+".description";
    struct stat st;
    if(stat(fname.c_str(),&st) == 0) return false;
    int h = ::open(fname.c_str(),O_RDWR | O_CREAT | O_EXCL,0600);
    // So far assume control directory is on local fs.
    // TODO: add locks or links for NFS
    if(h == -1) return false;
    fix_file_owner(fname,*user);
    close(h);
    break;
  };
  if(i>=100) { job_id=""; return false; };
  return true;
}

/*
  name - name of file to access
  locked - return false if job already running
  jobid - returns id extracted from name
  logname - name of log file (errors, status, etc.)
  log - stdlog of job
  spec_dir - if file belogs to virtual directory 
*/
int JobPlugin::is_allowed(const char* name,bool locked,bool* spec_dir,std::string* jobid,char const ** logname,std::string* log) {
  if(logname) (*logname) = NULL;
  if(log) (*log)="";
  if(spec_dir) (*spec_dir)=false;
  JobId id(name);
  if(id == "info") {
    if(spec_dir) (*spec_dir)=false;
    return (IS_ALLOWED_READ | IS_ALLOWED_LIST);
  };
  if(strncmp(id.c_str(),"info/",5) == 0) {
    if(spec_dir) (*spec_dir)=true;
    name+=5; id=name;
    int n=id.find('/'); if(n != std::string::npos) id.erase(n);
    if(jobid) (*jobid)=id;
    if(id.length() == 0) return 0;
    const char* l_name = name+id.length();
    if(l_name[0] == '/') l_name++;
    if(logname) { (*logname)=l_name; };
    JobLocalDescription job_desc;
    int res = 0;
    if(!job_local_read_file(id,*user,job_desc)) return false;
    if(job_desc.DN != subject) {
      // Not an owner. Check acl.
      if(user_a) {
        std::string acl_file = user->ControlDir()+"/job."+id+".acl";
        struct stat st;
        if(stat(acl_file.c_str(),&st) == 0) {
          if(S_ISREG(st.st_mode)) {
            GACLacl* acl = GACLloadAcl((char*)(acl_file.c_str()));
            if(acl) {
              GACLperm perm = AuthUserGACLTest(acl,*user_a);
              int res = 0;
              if(GACLhasList(perm))
                res|=IS_ALLOWED_LIST;
              if(GACLhasRead(perm) | GACLhasWrite(perm))
                res|=(IS_ALLOWED_READ | IS_ALLOWED_LIST);
              if(GACLhasAdmin(perm))
                res|=(IS_ALLOWED_READ | IS_ALLOWED_WRITE | IS_ALLOWED_LIST);
              return res;
            };
          };
        };
      };
      return 0;
    };
    if(strncmp(l_name,"proxy",5) == 0) return (IS_ALLOWED_LIST);
    return (IS_ALLOWED_READ | IS_ALLOWED_WRITE | IS_ALLOWED_LIST);
  };
  int n=id.find('/'); if(n != std::string::npos) id.erase(n);
  if(jobid) (*jobid)=id;
  JobLocalDescription job_desc;
  if(job_local_read_file(id,*user,job_desc)) {
    int res = 0;
    bool spec = false;
    if(log) (*log)=job_desc.stdlog;
    if(n != std::string::npos) {
      int l = job_desc.stdlog.length();
      if(l != 0) {
        if(strncmp(name+n+1,job_desc.stdlog.c_str(),l) == 0) {
          if(name[n+1+l] == 0) {
            if(spec_dir) (*spec_dir)=true;
            if(logname) (*logname)=name+n+1+l;
            spec=true;
          };
          if(name[n+1+l] == '/') {
            if(spec_dir) (*spec_dir)=true;
            if(logname) (*logname)=name+n+1+l+1;
            spec=true;
          };
        };
      };
    };
    if(job_desc.DN == subject) {
      res|=(IS_ALLOWED_READ | IS_ALLOWED_WRITE | IS_ALLOWED_LIST);
    } else {
      // Not an owner. Check acl.
      if(user_a) {
        std::string acl_file = user->ControlDir()+"/job."+id+".acl";
        struct stat st;
        if(stat(acl_file.c_str(),&st) == 0) {
          if(S_ISREG(st.st_mode)) {
            GACLacl* acl = GACLloadAcl((char*)(acl_file.c_str()));
            if(acl) {
              GACLperm perm = AuthUserGACLTest(acl,*user_a);
              if(spec) {
                if(GACLhasList(perm))
                  res|=IS_ALLOWED_LIST;
                if(GACLhasRead(perm) | GACLhasWrite(perm))
                  res|=(IS_ALLOWED_READ | IS_ALLOWED_LIST);
                if(GACLhasAdmin(perm))
                  res|=(IS_ALLOWED_READ | IS_ALLOWED_WRITE | IS_ALLOWED_LIST);
              } else {
                if(GACLhasList(perm)) res|=IS_ALLOWED_LIST;
                if(GACLhasRead(perm)) res|=IS_ALLOWED_READ;
                if(GACLhasWrite(perm)) res|=IS_ALLOWED_WRITE;
                if(GACLhasAdmin(perm))
                  res|=(IS_ALLOWED_READ | IS_ALLOWED_WRITE | IS_ALLOWED_LIST);
              };
              return res;
            };
          };
        };
      };
    };
    if(res) {
      if(!locked) return res;
      job_state_t status=job_state_read_file(id,*user);
      if((status == JOB_STATE_ACCEPTED) || (status == JOB_STATE_PREPARING)) {
        return res; 
      };
    };
  };
  return 0;
}

